import { __awaiter, __generator } from "tslib";
import { version } from "./../../package.json";
import { requestIdleCallbackIfAvailable } from "./utils/async";
import { x64hash128 } from "./utils/hashing";
import { errorToObject } from "./utils/misc";
import getConfidence from "./confidence";
import getFonts from "./sources/fonts";
import getCanvasFingerprint from "./sources/canvas";
import getDomBlockers from "./sources/dom_blockers";
import getFontPreferences from "./sources/font_preferences";
import getOsCpu from "./sources/os_cpu";
import getLanguages from "./sources/languages";
import getColorDepth from "./sources/color_depth";
import getDeviceMemory from "./sources/device_memory";
import getScreenResolution from "./sources/screen_resolution";
import getHardwareConcurrency from "./sources/hardware_concurrency";
import getTimezone from "./sources/timezone";
import getSessionStorage from "./sources/session_storage";
import getLocalStorage from "./sources/local_storage";
import getIndexedDB from "./sources/indexed_db";
import getOpenDatabase from "./sources/open_database";
import getCpuClass from "./sources/cpu_class";
import getPlatform from "./sources/platform";
import getPlugins from "./sources/plugins";
import getTouchSupport from "./sources/touch_support";
import getVendor from "./sources/vendor";
import getVendorFlavors from "./sources/vendor_flavors";
import areCookiesEnabled from "./sources/cookies_enabled";
import getColorGamut from "./sources/color_gamut";
import areColorsInverted from "./sources/inverted_colors";
import areColorsForced from "./sources/forced_colors";
import getMonochromeDepth from "./sources/monochrome";
import getContrastPreference from "./sources/contrast";
import isMotionReduced from "./sources/reduced_motion";
import isHDR from "./sources/hdr";
import getMathFingerprint from "./sources/math";
function componentsToCanonicalString(components) {
    var result = "";
    for (var _i = 0, _a = Object.keys(components).sort(); _i < _a.length; _i++) {
        var componentKey = _a[_i];
        var component = components[componentKey];
        var value = component.error ? "error" : JSON.stringify(component.value);
        result += "" + (result ? "|" : "") + componentKey.replace(/([:|\\])/g, "\\$1") + ":" + value;
    }
    return result;
}
export function componentsToDebugString(components) {
    return JSON.stringify(components, function (_key, value) {
        if (value instanceof Error) {
            return errorToObject(value);
        }
        return value;
    }, 2);
}
export function hashComponents(components) {
    // return x64hash128(componentsToCanonicalString(components));
    return componentsToCanonicalString(components);
}
/**
 * Makes a GetResult implementation that calculates the visitor id hash on demand.
 * Designed for optimisation.
 */
function makeLazyGetResult(components) {
    var visitorIdCache;
    // This function runs very fast, so there is no need to make it lazy
    var confidence = getConfidence(components);
    // A plain class isn't used because its getters and setters aren't enumerable.
    return {
        get visitorId() {
            if (visitorIdCache === undefined) {
                visitorIdCache = hashComponents(this.components);
            }
            return visitorIdCache;
        },
        set visitorId(visitorId) {
            visitorIdCache = visitorId;
        },
        confidence: confidence,
        components: components,
        version: version,
    };
}
/**
 * A delay is required to ensure consistent entropy components.
 * See https://github.com/fingerprintjs/fingerprintjs/issues/254
 * and https://github.com/fingerprintjs/fingerprintjs/issues/307
 * and https://github.com/fingerprintjs/fingerprintjs/commit/945633e7c5f67ae38eb0fea37349712f0e669b18
 */
export function prepareForSources(delayFallback) {
    if (delayFallback === void 0) { delayFallback = 50; }
    // A proper deadline is unknown. Let it be twice the fallback timeout so that both cases have the same average time.
    return requestIdleCallbackIfAvailable(delayFallback, delayFallback * 2);
}
/**
 * The function isn't exported from the index file to not allow to call it without `load()`.
 * The hiding gives more freedom for future non-breaking updates.
 *
 * A factory function is used instead of a class to shorten the attribute names in the minified code.
 * Native private class fields could've been used, but TypeScript doesn't allow them with `"target": "es5"`.
 */
function makeAgent(getComponents, debug) {
    var creationTime = Date.now();
    return {
        get: function (options) {
            return __awaiter(this, void 0, void 0, function () {
                var startTime, components, result;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            startTime = Date.now();
                            return [4 /*yield*/, getComponents()];
                        case 1:
                            components = _a.sent();
                            result = makeLazyGetResult(components);
                            if (debug || (options === null || options === void 0 ? void 0 : options.debug)) {
                                // console.log is ok here because it's under a debug clause
                                // eslint-disable-next-line no-console
                                console.log("Copy the text below to get the debug data:\n\n```\nversion: " + result.version + "\nuserAgent: " + navigator.userAgent + "\ntimeBetweenLoadAndGet: " + (startTime - creationTime) + "\nvisitorId: " + result.visitorId + "\ncomponents: " + componentsToDebugString(components) + "\n```");
                            }
                            return [2 /*return*/, result];
                    }
                });
            });
        },
    };
}
/**
 * Sends an unpersonalized AJAX request to collect installation statistics
 */
/**
 * Builds an instance of Agent and waits a delay required for a proper operation.
 */
// export async function load({
//   delayFallback,
//   debug,
//   monitoring = true,
// }: Readonly<LoadOptions> = {}): Promise<Agent> {
//   if (monitoring) {
//     monitor();
//   }
//   await prepareForSources(delayFallback);
//   const getComponents = loadBuiltinSources({ debug });
//   return makeAgent(getComponents, debug);
// }
var sources = {
    // READ FIRST:
    // See https://github.com/fingerprintjs/fingerprintjs/blob/master/contributing.md#how-to-make-an-entropy-source
    // to learn how entropy source works and how to make your own.
    // The sources run in this exact order.
    // The asynchronous sources are at the start to run in parallel with other sources.
    fonts: getFonts,
    domBlockers: getDomBlockers,
    fontPreferences: getFontPreferences,
    // audio: getAudioFingerprint,
    // screenFrame: getRoundedScreenFrame,
    osCpu: getOsCpu,
    languages: getLanguages,
    colorDepth: getColorDepth,
    deviceMemory: getDeviceMemory,
    screenResolution: getScreenResolution,
    hardwareConcurrency: getHardwareConcurrency,
    timezone: getTimezone,
    sessionStorage: getSessionStorage,
    localStorage: getLocalStorage,
    indexedDB: getIndexedDB,
    openDatabase: getOpenDatabase,
    cpuClass: getCpuClass,
    platform: getPlatform,
    plugins: getPlugins,
    canvas: getCanvasFingerprint,
    touchSupport: getTouchSupport,
    vendor: getVendor,
    vendorFlavors: getVendorFlavors,
    cookiesEnabled: areCookiesEnabled,
    colorGamut: getColorGamut,
    invertedColors: areColorsInverted,
    forcedColors: areColorsForced,
    monochrome: getMonochromeDepth,
    contrast: getContrastPreference,
    reducedMotion: isMotionReduced,
    hdr: isHDR,
    math: getMathFingerprint,
};
var getVisitorId = function () { return __awaiter(void 0, void 0, void 0, function () {
    var components, loadArray, loadArrayPromise, visitorId;
    return __generator(this, function (_a) {
        switch (_a.label) {
            case 0:
                components = {};
                loadArray = Object.entries(sources);
                loadArrayPromise = Promise.all(loadArray.map(function (_a) {
                    var key = _a[0], value = _a[1];
                    return __awaiter(void 0, void 0, void 0, function () {
                        var result;
                        return __generator(this, function (_b) {
                            switch (_b.label) {
                                case 0: return [4 /*yield*/, value()];
                                case 1:
                                    result = _b.sent();
                                    components[key] = result;
                                    return [2 /*return*/, result];
                            }
                        });
                    });
                }));
                return [4 /*yield*/, loadArrayPromise];
            case 1:
                _a.sent();
                console.log("components", components);
                visitorId = x64hash128(JSON.stringify(components));
                console.log("visitorId", visitorId);
                return [2 /*return*/, visitorId];
        }
    });
}); };
export default getVisitorId;
